# -*- coding: utf-8 -*-
import math
import random

import pygame
from pygame.locals import *

from director import Scene, Node
import euclid
import data
import gdraw
import music
import colorsys


def v(o):
    return euclid.Point2(o.x, o.y)


class Weapon(Node):
    radius = 10
    color = (217, 91, 67)
    base_color = (217, 91, 67)

    def draw(self, target):
        sx, sy = self.to_screen((0, 0))
        pygame.draw.circle(target, self.color,
            (sx, sy), self.radius)


class Head(Node):
    radius = 10

    def draw(self, target):
        sx, sy = self.to_screen((0, 0))
        pygame.draw.circle(target, (255, 225, 130),
            (sx, sy), self.radius)


class HealthBar(Node):
    z = 10
    color = 192, 41, 66

    def draw(self, target):
        h = (self.parent.director.size[1] - 20) * \
            self.parent.player.energy / 100
        r = Rect(10, self.parent.director.size[1] - 10, 10, -h)
        pygame.draw.rect(target, self.color, r)


class ComboBar(Node):
    z = 10
    color = 83, 119, 122

    def draw(self, target):
        h = (self.parent.director.size[1] - 20) * \
            self.parent.player.combo / 100
        r = Rect(30, self.parent.director.size[1] - 10, 10, -h)
        pygame.draw.rect(target, self.color, r)


class ScoreIndicator(Node):
    z = 10
    color = 83, 119, 122
    size = 30

    def draw(self, target):
        font = pygame.font.Font(data.filepath("Fishfingers.ttf"),
            int(self.size))
        t = font.render(str(int(self.parent.player.score)), True, self.color)
        r = t.get_rect()
        x, y = self.parent.director.size
        r.bottom = y - 10
        r.right = x - 10
        target.blit(t, r)

    def update(self, dt):
        self.size = max(30, self.size - 20 * dt)


def to_one(x, y, z):
    return x / 255., y / 255., z / 255.


def to_short(x, y, z):
    return int(x * 255), int(y * 255), int(z * 255)


class Player(Node):
    color = 236, 208, 120

    def __init__(self):
        super(Player, self).__init__()
        self.weapon = None
        self.weapon_distance = 60
        self.max_velocity = 200
        self.radius = 30
        self.power = 20
        self.killed_her = False

        self.w_dir = 1
        self.w_vel = 5
        self.w_pos = 0
        self.w_rest_d = 20
        self.head = Head()
        self.head.x = 0
        self.head.y = self.radius - self.head.radius
        self.head.z = -1
        self.add(self.head)
        self.rotate = 0

        self.image = pygame.image.load(data.load("player.png"))

        self.reset()

    def reset(self):
        self.score = 0
        self.kills = 0
        self.fire_left = 0
        self.velocity = euclid.Vector2(0, 0)
        self.energy = 100
        self.combo = 0
        self.enemy_target = None
        self.firing = False

    def draw(self, target):
        p = self.to_screen((0, 0))
        #pygame.draw.circle(target, self.color,
            #p, self.radius)
        target.blit(self.image, euclid.Point2(*p) -
            (self.image.get_width() / 2, self.image.get_width() / 2))

    def apply_move(self, dt, extra=None):
        if extra is None:
            extra = euclid.Vector2(0, 0)

        self.velocity *= 0.2 ** dt
        pos = v(self) + (self.velocity + extra) * dt
        self.x = pos[0]
        self.y = pos[1]

    def update(self, dt):
        # XXCOMBO
        # if fire left and target;
        # if distance to target > X:
        # apply extra vel torward target (depending on combo)

        if self.weapon is not None:
            self.weapon.radius = int(10 + self.combo / 10.)
            h, s, value = colorsys.rgb_to_hsv(
                *to_one(*self.weapon.base_color))
            dc = 1 - value
            value += dc * (self.combo / 100.)
            dc = 1 - s
            s += dc * (self.combo / 100.)
            self.weapon.color = to_short(*colorsys.hsv_to_rgb(
                h, s, value))

        if self.firing and self.fire_left <= 0:
            self.combo = 0

        if self.fire_left and self.enemy_target is not None:
            enemy_v = v(self.enemy_target) - v(self)
            if abs(enemy_v) > self.enemy_target.radius + 80:
                extra = enemy_v.normalized() * (300 + 6 * self.combo)
                self.apply_move(dt, extra)
            else:
                self.enemy_target = None
                self.velocity = euclid.Vector2(0, 0)
                self.apply_move(dt)
        else:
            if self.enemy_target is not None:
                self.enemy_target = None
                self.velocity = euclid.Vector2(0, 0)
                # fire timeout with target!
                self.combo = 0
            self.apply_move(dt)

        if self.weapon is not None:
            if self.fire_left > 0:
                if self.fire_left > 0.5:
                    self.weapon_distance = self.w_rest_d + \
                        math.sin(self.fire_left * math.pi) ** 0.35 * 60
                else:
                    self.weapon_distance = self.w_rest_d + \
                        math.sin(self.fire_left * math.pi) ** 3 * 60
            else:
                self.weapon_distance = self.w_rest_d

            aim = v(self.head)
            pos = aim.normalized() * self.weapon_distance
            self.weapon.x = pos.x
            self.weapon.y = pos.y

        aim = v(self.head)
        pos = euclid.Matrix3.new_rotate(-6 * dt * self.rotate) * aim

        self.head.x = pos.x
        self.head.y = pos.y

        cdt = self.combo * 0.2
        self.combo -= cdt * dt

        if self.fire_left > 0:
            self.fire_left -= dt

    def accel(self, ax, ay):
        av = euclid.Vector2(ax, ay)
        nv = self.velocity + av * self.power
        if abs(self.velocity) > self.max_velocity:
            # dont increase velocity if over max velocity
            if abs(nv) <= abs(self.velocity):
                self.velocity = nv
        else:
            # increase up to max_velocity if under
            if abs(nv) <= self.max_velocity:
                self.velocity = nv
            else:
                self.velocity = nv.normalized() * self.max_velocity

    def walk(self, ax, ay):
        aim = v(self.head)
        forward = -aim.normalized()
        final = forward * ay
        final.normalize()
        nv = self.velocity + final * self.power

        if abs(self.velocity) > self.max_velocity:
            # dont increase velocity if over max velocity
            if abs(nv) <= abs(self.velocity):
                self.velocity = nv
        else:
            # increase up to max_velocity if under
            if abs(nv) <= self.max_velocity:
                self.velocity = nv
            else:
                self.velocity = nv.normalized() * self.max_velocity

        self.rotate = ax

    def fire(self):
        # select the target to aim for
        self.firing = True
        enemies_left = [x for x in self.parent.children
                        if isinstance(x, Enemy)]
        aim = v(self.head)
        p = v(self)
        self.enemy_target = None
        t_dist = float("inf")
        for enemy in enemies_left:
            enemy_v = v(enemy) - p
            try:
                angle = abs(aim.angle(enemy_v))
            except ValueError:
                angle = 0

            if angle < math.radians(20):
                if abs(enemy_v) < t_dist:
                    t_dist = abs(enemy_v)
                    self.enemy_target = enemy

        if self.enemy_target is not None:
            head_v = abs(v(self.head)) * \
                (v(self.enemy_target) - v(self)).normalized()
            self.head.x = head_v.x
            self.head.y = head_v.y

        self.fire_left = 1

    def damage(self, damage):
        self.parent.director.sounds.self_hit.play()
        self.energy -= damage
        if self.energy <= 0:
            self.parent.die()

    def hit(self):
        self.combo += 10
        self.combo = min(100, self.combo)
        self.score += self.combo
        self.parent.score.size = self.combo


class Enemy(Node):
    collides = True
    radius = 20
    energy = 3
    speed = 30
    damage = 10
    last_hit_timeout = 0
    last_hit = None
    velocity = euclid.Vector2(0, 0)
    power = 10
    max_velocity = 100
    color = 217, 91, 67
    pushback_value = 125
    sprite = "enemy.png"
    image = None

    def accel(self, ax, ay):
        av = euclid.Vector2(ax, ay)
        nv = self.velocity + av * self.power
        if abs(self.velocity) > self.max_velocity:
            # dont increase velocity if over max velocity
            if abs(nv) <= abs(self.velocity):
                self.velocity = nv
        else:
            # increase up to max_velocity if under
            if abs(nv) <= self.max_velocity:
                self.velocity = nv
            else:
                self.velocity = nv.normalized() * self.max_velocity

    def draw(self, target):
        if self.image is None:
            self.image = pygame.image.load(data.load(self.sprite))
        p = self.to_screen((0, 0))
        target.blit(self.image, euclid.Point2(*p) -
            (self.image.get_width() / 2, self.image.get_width() / 2))
        #pygame.draw.circle(target, self.color,
        #    self.to_screen((0, 0)), self.radius)

    def collision(self, scene, other):
        if self.last_hit == other and self.last_hit_timeout > 0:
            return

        self.last_hit = other
        self.last_hit_timeout = 0.2
        combo = self.parent.player.combo
        damaged = False
        if isinstance(other, Player):
            push = 300
            other.damage(self.damage)

            d = v(scene.player) - v(self)
            scene.player.velocity = euclid.Vector2(*d).normalized() * push
        else:
            self.parent.director.sounds.hit.play()
            self.energy -= 1 + 2 * combo / 100.

            self.pushback(other)

            self.parent.player.fire_left = 0
            self.parent.player.firing = False
            if self.parent.player.enemy_target is not None:
                self.parent.player.enemy_target = None
                self.parent.player.velocity = euclid.Vector2(0, 0)

            damaged = True
            self.parent.player.hit()

        if self.energy < 0:
            self.parent.director.sounds.kill.play()
            root = self.parent
            self.parent.remove(self)

            enemies_left = [x for x in root.children if isinstance(x, Enemy)]
            if not enemies_left:
                root.remove(root.barrier)

            # add powerup
            placed = False
            xs, ys = root.director.size
            vs = v(self)
            while not placed:
                x = (random.random() - 0.5) * (xs / 2 - 20)
                y = (random.random() - 0.5) * 2 * (ys / 2 - 20)
                for other in root.children:
                    if abs(vs - v(other)) < 20:
                        break
                else:
                    placed = True
                    p = Powerup()
                    p.x = x
                    p.y = y
                    root.add(p)

        elif damaged:

            self.parent.director.sounds.hit.set_volume(0.5 + combo / 100.)
            self.parent.director.sounds.hit.play()

    def pushback(self, other):
        ss = self.parent.player.to_screen(v(other))
        w = euclid.Vector2(*ss)
        sd = self.parent.to_screen(v(self))
        s = euclid.Vector2(*sd)
        d = s - w
        self.velocity = d.normalized() * self.pushback_value

    def update(self, dt):
        p = v(self.parent.player)
        s = v(self)
        d = p - s
        self.accel(*d)
        self.velocity *= 0.5 ** dt
        pos = v(self) + self.velocity * dt
        self.x = pos[0]
        self.y = pos[1]

        self.last_hit_timeout -= dt


class HeavyEnemy(Enemy):
    collides = True
    radius = 30
    energy = 30
    speed = 15
    damage = 10
    last_hit_timeout = 0
    last_hit = None
    velocity = euclid.Vector2(0, 0)
    power = 10
    max_velocity = 75
    color = 217, 91, 67
    pushback_value = 125
    sprite = "heavy_enemy.png"


class RamEnemy(Enemy):
    collides = True
    radius = 20
    energy = 15
    speed = 30
    damage = 10
    last_hit_timeout = 0
    last_hit = None
    velocity = euclid.Vector2(0, 0)
    power = 20
    max_velocity = 200
    color = 156, 29, 35
    pushback_value = 400
    sprite = "ram_enemy.png"


class Boss(Enemy):
    collides = True
    radius = 60
    energy = 75
    speed = 15
    damage = 10
    last_hit_timeout = 0
    last_hit = None
    velocity = euclid.Vector2(0, 0)
    power = 10
    max_velocity = 100
    color = 236, 208, 120
    pushback_value = 125
    fire_interval = 4
    last_fire = 0
    sprite = "boss.png"

    def update(self, dt):
        if self.fire_interval > 0:
            self.last_fire += dt
            if self.last_fire > self.fire_interval:
                self.last_fire = 0
                b = Bullet()
                b.x = self.x
                b.y = self.y
                b.z = 3
                self.parent.add(b)
        super(Boss, self).update(dt)

    def draw(self, target):
        if self.image is None:
            self.image = pygame.image.load(data.load(self.sprite))
        p = self.to_screen((0, 0))
        target.blit(self.image, euclid.Point2(*p) -
            (self.image.get_width() / 2, self.image.get_width() / 2))

        pygame.draw.circle(target, self.color,
            self.to_screen((0, 0)), int(self.radius / 2. * self.energy / 70))


class Boss2(Enemy):
    collides = True
    radius = 50
    speed = 15
    color = 217, 91, 67
    pushback_value = 125
    fire_interval = 10
    last_fire = 0
    sprite = "boss2.png"

    def update(self, dt):
        pass

    def collision(self, scene, other):
        if isinstance(other, Player):
            push = 300
            d = v(scene.player) - v(self)
            scene.player.velocity = euclid.Vector2(*d).normalized() * push
        else:
            self.parent.director.sounds.hit.play()
            self.parent.player.killed_her = True
            self.parent.remove(self)


class Bullet(Node):
    collides = True
    radius = 10
    energy = 3
    speed = 30
    damage = 50
    last_hit_timeout = 0
    last_hit = None
    velocity = 400
    power = 10
    max_velocity = 400
    color = 255, 255, 255
    pushback_value = 125
    direction = None
    sprite = "bullet.png"
    image = None

    def update(self, dt):
        if self.direction is None:
            p = v(self.parent.player)
            s = v(self)
            self.direction = p - s
        d = self.direction
        self.velocity += self.velocity * 0.8 * dt
        pos = d.normalized() * self.velocity * dt
        self.x += pos[0]
        self.y += pos[1]

        p = self.parent.to_screen((self.x, self.y))
        sx, sy = self.parent.director.size
        if p[0] < 0 or p[0] > sx or p[1] < 0 or p[1] > sy:
            self.parent.remove(self)

    def collision(self, scene, other):
        if isinstance(other, Player):
            push = 300
            other.damage(self.damage)

            d = v(scene.player) - v(self)
            scene.player.velocity = euclid.Vector2(*d).normalized() * push
            self.parent.player.fire_left = 0
            self.parent.player.combo = 0

            self.parent.director.sounds.kill.play()
            self.parent.remove(self)

    def draw(self, target):
        if self.image is None:
            self.image = pygame.image.load(data.load(self.sprite))
        p = self.to_screen((0, 0))
        target.blit(self.image, euclid.Point2(*p) -
            (self.image.get_width() / 2, self.image.get_width() / 2))
        #pygame.draw.circle(target, self.color,
        #    self.to_screen((0, 0)), self.radius)


class Goal(Node):
    image = None

    def draw(self, target):
        if self.image is None:
            self.image = pygame.image.load(data.load("goal.png"))

        p = self.to_screen((0, 0))
        target.blit(self.image, euclid.Point2(*p) -
            (self.image.get_width() / 2, self.image.get_width() / 2))


class WeaponPowerup(Node):
    radius = 20
    collides = True
    sprite = "powerup.png"
    image = None

    def draw(self, target):
        if self.image is None:
            self.image = pygame.image.load(data.load(self.sprite))
        p = self.to_screen((0, 0))
        target.blit(self.image, euclid.Point2(*p) -
            (self.image.get_width() / 2, self.image.get_width() / 2))

    def collision(self, scene, other):
        if isinstance(other, Player):
            self.parent.player.weapon = Weapon()
            self.parent.player.add(self.parent.player.weapon)
            self.parent.director.sounds.pickup.play()
        self.parent.remove(self)


class Powerup(Node):
    radius = 20
    collides = True
    sprite = "powerup.png"
    image = None
    time_left = 5

    def draw(self, target):
        if self.image is None:
            self.image = pygame.image.load(data.load(self.sprite))
        p = self.to_screen((0, 0))
        target.blit(self.image, euclid.Point2(*p) -
            (self.image.get_width() / 2, self.image.get_width() / 2))

    def collision(self, scene, other):
        if self.parent:
            self.parent.player.score += 500
            self.parent.player.combo = 100
            self.parent.player.firing = False
            self.parent.director.sounds.powerup.play()
            self.parent.remove(self)

    def update(self, dt):
        self.time_left -= dt
        if self.time_left < 0:
            self.parent.director.sounds.powerdown.play()
            self.parent.remove(self)


class KeyBarrier(Node):
    radius = 40
    collides = True
    energy = 5
    damage = 5
    sprite = "barrier.png"
    image = None

    def draw(self, target):
        if self.image is None:
            self.image = pygame.image.load(data.load(self.sprite))

        p = self.to_screen((0, 0))
        target.blit(self.image, euclid.Point2(*p) -
            (self.image.get_width() / 2, self.image.get_width() / 2))

    def collision(self, scene, other):
        push = 400

        if isinstance(other, Player):
            other.damage(self.damage)

        d = v(scene.player) - v(self)
        scene.player.velocity = euclid.Vector2(*d).normalized() * push


class Barrier(Node):
    radius = 40
    collides = True
    energy = 5
    damage = 5
    sprite = "barrier.png"
    image = None

    def draw(self, target):
        if self.image is None:
            self.image = pygame.image.load(data.load(self.sprite))

        p = self.to_screen((0, 0))
        target.blit(self.image, euclid.Point2(*p) -
            (self.image.get_width() / 2, self.image.get_width() / 2))

    def collision(self, scene, other):
        if isinstance(other, Player):
            push = 400
            other.damage(self.damage)
            d = v(scene.player) - v(self)
            scene.player.velocity = euclid.Vector2(*d).normalized() * push
        else:
            self.parent.director.sounds.hit.play()
            self.energy -= self.parent.player.fire_left
            #self.parent.player.fire_left = 0
            self.parent.player.w_dir *= -1

        if self.energy < 0:
            self.parent.remove(self)


class Title(Scene):

    def __init__(self):
        super(Title, self).__init__()
        self.inited = False

    def draw(self, target):
        if not self.inited:
            self.inited = True
            self.director.sounds.intro.play(-1)
            self.image = pygame.image.load(data.load("title.png"))
            self.bg = pygame.image.load(data.load("bg.png"))

            target.blit(self.bg, (0, 0))
            target.blit(self.image, (0, 0))

            font = pygame.font.Font(data.filepath("Fishfingers.ttf"), 60)
            t = font.render("press space to begin", True,
                (83, 119, 122))
            r = t.get_rect()
            x, y = self.director.size
            r.centery = y / 3 * 2
            r.centerx = x / 2
            target.blit(t, r)

    def event(self, ev):
        if ev.type == KEYDOWN:
            if ev.key == K_SPACE:
                self.director.sounds.intro.stop()
                self.director.replace(Intro(None))


class CutScene(Scene):
    is_final = False

    def __init__(self, last_scene=None):
        super(CutScene, self).__init__()
        self.inited = False
        self.since_start = 0
        if last_scene is not None:
            self.last_scene = last_scene
        self.nokeys = 1
        self.readkeys = False

    def draw(self, target):
        if not self.inited:
            self.inited = True
            if not self.is_final:
                self.director.sounds.cut.play(-1, fade_ms=1000)
            else:
                self.director.sounds.victory.play()
            self.bg = pygame.image.load(data.load("bg.png"))

            target.blit(self.bg, (0, 0))

            font = pygame.font.Font(data.filepath("Brain Flower.ttf"), 60)
            ys = 100
            for line in self.lines:
                t = font.render(line, True, (83, 119, 122))

                r = t.get_rect()
                x, _ = self.director.size
                r.centery = ys
                r.centerx = x / 2
                target.blit(t, r)
                ys += 80

        if not self.readkeys and self.since_start > self.nokeys:
            font = pygame.font.Font(data.filepath("Brain Flower.ttf"), 25)
            t = font.render("[press space to continue]",
                True, (83, 119, 122))

            r = t.get_rect()
            x, y = self.director.size
            r.bottom = y - 10
            r.centerx = x / 2
            target.blit(t, r)
            self.readkeys = True

    def update(self, dt):
        self.since_start += dt

    def event(self, ev):
        if self.readkeys:
            if ev.type == KEYDOWN:
                if ev.key == K_SPACE:
                    self.director.sounds.cut.fadeout(1000)
                    if hasattr(self, "last_scene"):
                        self.director.replace(self.next_scene(self.last_scene))
                    else:
                        self.director.replace(Level0())


class Level0(Scene):
    fire_interval = 0

    def __init__(self):
        super(Level0, self).__init__()
        self.setup_player()
        self.setup_level()
        self.music_setup_done = False

        self.x_accel = 0
        self.y_accel = 0
        self.last_fire = 0
        self.start_x = self.start_y = 0
        self.add(HealthBar())
        self.add(ComboBar())
        self.score = ScoreIndicator()
        self.add(self.score)
        self.accum = None
        self.inited = False

    def setup_level(self):
        self.goal = Goal()
        self.goal.x, self.goal.y = 300, 300
        self.goal.z = -2
        self.add(self.goal)

    def setup_player(self):
        self.player = Player()
        self.add(self.player)

    def render(self, target):
        if not self.inited:
            self.inited = True
            self.bg = pygame.image.load(data.load("bg.png"))

        screen = target
        if self.accum is None:
            self.accum = \
                gdraw.new(target.get_width(), target.get_height())

        target = self.accum
        gdraw.fade(target)

        self.draw(screen)
        self.goal.render(screen)
        for c in sorted(self.children, key=lambda x: x.z):
            if not isinstance(c, Goal):
                c.render(target)

        screen.blit(target, (0, 0))

    def draw(self, target):
        target.blit(self.bg, (0, 0))

    def event(self, ev):
        if ev.type == KEYDOWN:
            if ev.key == K_RIGHT:
                self.x_accel = 1
            elif ev.key == K_LEFT:
                self.x_accel = -1
            if ev.key == K_DOWN:
                self.y_accel = 1
            elif ev.key == K_UP:
                self.y_accel = -1
            elif ev.key == K_SPACE:
                self.player.fire()
            elif ev.key == K_SEMICOLON:
                self.goal_reached()
        elif ev.type == KEYUP:
            if ev.key == K_RIGHT:
                if self.x_accel == 1:
                    self.x_accel = 0
            elif ev.key == K_LEFT:
                if self.x_accel == -1:
                    self.x_accel = 0
            if ev.key == K_DOWN:
                if self.y_accel == 1:
                    self.y_accel = 0
            elif ev.key == K_UP:
                if self.y_accel == -1:
                    self.y_accel = 0

    def update(self, dt):
        if self.fire_interval > 0:
            self.last_fire += dt
            if self.last_fire > self.fire_interval:
                self.last_fire = 0
                b = Bullet()
                b.x = self.goal.x
                b.y = self.goal.y
                self.add(b)

        if not self.music_setup_done:
            self.setup_music()
            self.music_setup_done = True

        self.player.walk(-self.x_accel, self.y_accel)

        if self.goal is not None:
            g = v(self.goal)
            p = v(self.player)
            if abs(g - p) < 35:
                self.director.sounds.portal2.play()
                self.goal_reached()

        self.check_collisions()

        bx, by = self.director.size
        bx /= 2
        by /= 2

        if self.player.x < -bx or self.player.x > bx:
            self.die()
        elif self.player.y < -by or self.player.y > by:
            self.die()

        self.do_music()

    def goal_reached(self):
        self.director.replace(CutLevel1(self))

    def check_collisions(self):
        p = self.player

        for c in self.children:
            if p == c:
                continue

            if c.collides:
                vc = v(c)
                vp = v(p)

                if abs(vp - vc) < p.radius + c.radius:
                    c.collision(self, p)

                if p.weapon is not None:
                    vw = v(p) + v(p.weapon)
                    if abs(vw - vc) < p.weapon.radius + c.radius:
                        c.collision(self, p.weapon)

    def die(self):
        self.director.sounds.die.play()
        self.director.replace(Death(self.player, self))

    def setup_music(self):
        self.music = music.Music(self.director.framerate)

    def do_music(self):
        self.music.update(self.player.combo, self.player.energy < 40)


class Intro(CutScene):
    lines = [
        "Your name is Vam.",
        "Ormag, the god, is your enemy.",
        "Arrogance some say, but you have decided",
        "that it is time to end injustice in this world.",
        "A god can be no more if we are to be his play toys.",
        "",
        "Cross The Doors Of Perception and find him!"
        ]


class Death(Scene):
    inited = False
    lines = [
        "You died.",
        "",
        "A Note Of Enlightment for you, consider Idolatry."
        ]
    since_start = 0
    nokeys = 2
    readkeys = False

    def __init__(self, player, last):
        self.last = last
        super(Death, self).__init__()
        self.last = last

    def draw(self, target):
        if not self.inited:
            self.inited = True
            lines = self.lines + [
                "Your achieved karma is a miserable %s." %
                    (int(self.last.player.score),)
                    ]
            self.bg = pygame.image.load(data.load("bg.png"))

            target.blit(self.bg, (0, 0))

            font = pygame.font.Font(data.filepath("Brain Flower.ttf"), 60)
            ys = 100
            for line in lines:
                t = font.render(line, True, (83, 119, 122))

                r = t.get_rect()
                x, _ = self.director.size
                r.centery = ys
                r.centerx = x / 2
                target.blit(t, r)
                ys += 80

        if not self.readkeys and self.since_start > self.nokeys:
            font = pygame.font.Font(data.filepath("Brain Flower.ttf"), 30)
            t = font.render("[press space to retry or enter for main menu]",
                True, (83, 119, 122))

            r = t.get_rect()
            x, y = self.director.size
            r.bottom = y - 10
            r.centerx = x / 2
            target.blit(t, r)
            self.readkeys = True

    def event(self, ev):
        if self.readkeys:
            if ev.type == KEYDOWN:
                if ev.key == K_SPACE:
                    if self.last.__class__ == Level0:
                        self.director.replace(Level0())
                    else:
                        self.last.player.reset()
                        self.last.player.x = self.last.start_x
                        self.last.player.y = self.last.start_y

                        self.director.replace(
                            self.last.__class__(self.last)
                            )
                elif ev.key in [K_RETURN, K_ESCAPE]:
                    self.director.replace(Title())

    def update(self, dt):
        self.since_start += dt


class Level1(Level0):
    def __init__(self, prev):
        super(Level1, self).__init__()
        self.player = prev.player
        self.player.velocity = euclid.Point2(0, 0)
        self.add(prev.player)
        if self.player.weapon is not None:
            self.player.children.remove(self.player.weapon)
            self.player.weapon = None
        self.start_x = self.player.x
        self.start_y = self.player.y

    def setup_level(self):
        self.goal = Goal()
        self.goal.x, self.goal.y = -300, -300
        self.goal.z = -2
        self.add(self.goal)

        self.barrier = Barrier()
        self.barrier.x, self.barrier.y = -300, -300
        self.barrier.z = -1
        self.add(self.barrier)

        wpu = WeaponPowerup()
        wpu.x, wpu.y = 300, -300
        wpu.z = -1
        self.add(wpu)

    def setup_player(self):
        pass

    def goal_reached(self):
        self.director.replace(CutLevel2(self))


class CutLevel1(CutScene):
    lines = [
        "Heaven And Hell, so alike."
        "",
        "",
        "And still, someone decides to help you. Why?"
        ]

    next_scene = Level1


class Level2(Level0):
    def __init__(self, prev):
        super(Level2, self).__init__()
        self.player = prev.player
        self.player.velocity = euclid.Point2(0, 0)
        self.add(prev.player)
        self.start_x = self.player.x
        self.start_y = self.player.y

    def setup_level(self):
        self.goal = Goal()
        self.goal.x, self.goal.y = -300, 300
        self.goal.z = -2
        self.add(self.goal)

        self.barrier = KeyBarrier()
        self.barrier.x, self.barrier.y = -300, 300
        self.barrier.z = -1
        self.add(self.barrier)

        wpu = Enemy()
        wpu.x, wpu.y = 300, -300
        wpu.z = -1
        self.add(wpu)

    def setup_player(self):
        pass

    def goal_reached(self):
        self.director.replace(CutLevel3(self))


class CutLevel2(CutScene):
    lines = [
        "The Devils Of Loudun rise to stop you."
        "",
        "You cant stop now!"
        ]

    next_scene = Level2


class Level3(Level0):
    def __init__(self, prev):
        super(Level3, self).__init__()
        self.player = prev.player
        self.player.velocity = euclid.Point2(0, 0)
        self.add(prev.player)
        self.start_x = self.player.x
        self.start_y = self.player.y

    def setup_level(self):
        self.goal = Goal()
        self.goal.x, self.goal.y = 300, -300
        self.goal.z = -2
        self.add(self.goal)

        self.barrier = KeyBarrier()
        self.barrier.x, self.barrier.y = 300, -300
        self.barrier.z = -1
        self.add(self.barrier)

        wpu = Enemy()
        wpu.x, wpu.y = 300, 300
        wpu.z = -1
        self.add(wpu)

        wpu = Enemy()
        wpu.x, wpu.y = -300, -300
        wpu.z = -1
        self.add(wpu)

    def setup_player(self):
        pass

    def goal_reached(self):
        self.director.replace(CutLevel4(self))


class CutLevel3(CutScene):
    lines = [
        "",
        "",
        "Themes and Variations."
        ]

    next_scene = Level3


class Level4(Level0):
    def __init__(self, prev):
        super(Level4, self).__init__()
        self.player = prev.player
        self.player.velocity = euclid.Point2(0, 0)
        self.add(prev.player)
        self.start_x = self.player.x
        self.start_y = self.player.y

    def setup_level(self):
        self.goal = Goal()
        self.goal.x, self.goal.y = -300, 300
        self.goal.z = -2
        self.add(self.goal)

        self.barrier = KeyBarrier()
        self.barrier.x, self.barrier.y = -300, 300
        self.barrier.z = -1
        self.add(self.barrier)

        wpu = Enemy()
        wpu.x, wpu.y = 300, 300
        wpu.z = -1
        self.add(wpu)

        wpu = Enemy()
        wpu.x, wpu.y = -300, -300
        wpu.z = -1
        self.add(wpu)

        wpu = Enemy()
        wpu.x, wpu.y = -150, 150
        wpu.z = -1
        self.add(wpu)

    def setup_player(self):
        pass

    def goal_reached(self):
        self.director.replace(CutLevel5(self))


class CutLevel4(CutScene):
    lines = [
        "",
        "Now More Than Ever",
        "Knowledge And Understanding are your tools."
        ]

    next_scene = Level4


class Level5(Level0):
    def __init__(self, prev):
        super(Level5, self).__init__()
        self.player = prev.player
        self.player.velocity = euclid.Point2(0, 0)
        self.add(prev.player)
        self.start_x = self.player.x
        self.start_y = self.player.y

    def setup_level(self):
        self.goal = Goal()
        self.goal.x, self.goal.y = 300, -300
        self.goal.z = -2
        self.add(self.goal)

        self.barrier = KeyBarrier()
        self.barrier.x, self.barrier.y = 300, -300
        self.barrier.z = -1
        self.add(self.barrier)

        wpu = Enemy()
        wpu.x, wpu.y = 300, 300
        wpu.z = -1
        self.add(wpu)

        wpu = Enemy()
        wpu.x, wpu.y = -300, -300
        wpu.z = -1
        self.add(wpu)

        wpu = RamEnemy()
        wpu.x, wpu.y = 150, -150
        wpu.z = -1
        self.add(wpu)

    def setup_player(self):
        pass

    def goal_reached(self):
        self.director.replace(CutLevel6(self))


class CutLevel5(CutScene):
    lines = [
        "",
        "Pacifism And Philosophy will not work.",
        ""
        ]

    next_scene = Level5


class Level6(Level0):
    def __init__(self, prev):
        super(Level6, self).__init__()
        self.player = prev.player
        self.player.velocity = euclid.Point2(0, 0)
        self.add(prev.player)
        self.start_x = self.player.x
        self.start_y = self.player.y

    def setup_level(self):
        self.goal = Goal()
        self.goal.x, self.goal.y = -300, 300
        self.goal.z = -2
        self.add(self.goal)

        self.barrier = KeyBarrier()
        self.barrier.x, self.barrier.y = -300, 300
        self.barrier.z = -1
        self.add(self.barrier)

        wpu = RamEnemy()
        wpu.x, wpu.y = 300, 300
        wpu.z = -1
        self.add(wpu)

        wpu = RamEnemy()
        wpu.x, wpu.y = -300, -300
        wpu.z = -1
        self.add(wpu)

    def setup_player(self):
        pass

    def goal_reached(self):
        self.director.replace(CutLevel7(self))


class CutLevel6(CutScene):
    lines = [
        "",
        "Along the Road.",
        ""
        ]

    next_scene = Level6


class Level7(Level0):
    def __init__(self, prev):
        super(Level7, self).__init__()
        self.player = prev.player
        self.player.velocity = euclid.Point2(0, 0)
        self.add(prev.player)
        self.start_x = self.player.x
        self.start_y = self.player.y

    def setup_level(self):
        self.goal = Goal()
        self.goal.x, self.goal.y = 300, -300
        self.goal.z = -2
        self.add(self.goal)

        self.barrier = KeyBarrier()
        self.barrier.x, self.barrier.y = 300, -300
        self.barrier.z = -1
        self.add(self.barrier)

        wpu = HeavyEnemy()
        wpu.x, wpu.y = 300, 300
        wpu.z = -1
        self.add(wpu)

        wpu = RamEnemy()
        wpu.x, wpu.y = -300, -300
        wpu.z = -1
        self.add(wpu)

    def setup_player(self):
        pass

    def goal_reached(self):
        self.director.replace(CutLevel8(self))


class CutLevel7(CutScene):
    lines = [
        "",
        "Art and Religion.",
        "",
        "Words and Reality.",
        "",
        "Fast and slow.",
        ""
        ]

    next_scene = Level7


class Level8(Level0):
    fire_interval = 5

    def __init__(self, prev):
        super(Level8, self).__init__()
        self.player = prev.player
        self.player.velocity = euclid.Point2(0, 0)
        self.add(prev.player)
        self.start_x = self.player.x
        self.start_y = self.player.y

    def setup_level(self):
        self.goal = Goal()
        self.goal.x, self.goal.y = -300, 300
        self.goal.z = -2
        self.add(self.goal)

        self.barrier = KeyBarrier()
        self.barrier.x, self.barrier.y = -300, 300
        self.barrier.z = -1
        self.add(self.barrier)

        wpu = HeavyEnemy()
        wpu.x, wpu.y = 300, 300
        wpu.z = -1
        self.add(wpu)

    def setup_player(self):
        pass

    def goal_reached(self):
        self.director.replace(CutLevelBoss(self))


class CutLevel8(CutScene):
    lines = [
        "",
        "Beware of distractions.",
        ""
        ]

    next_scene = Level8


class LevelBoss(Level0):
    def __init__(self, prev):
        super(LevelBoss, self).__init__()
        self.player = prev.player
        self.player.velocity = euclid.Point2(0, 0)
        self.add(prev.player)
        self.start_x = self.player.x
        self.start_y = self.player.y

    def setup_level(self):
        self.goal = Goal()
        self.goal.x, self.goal.y = 300, -300
        self.goal.z = -2
        self.add(self.goal)

        self.barrier = KeyBarrier()
        self.barrier.x, self.barrier.y = 300, -300
        self.barrier.z = -1
        self.add(self.barrier)

        wpu = Boss()
        wpu.x, wpu.y = 300, 300
        wpu.z = -1
        self.add(wpu)

    def setup_player(self):
        pass

    def goal_reached(self):
        self.director.replace(CutLevelBoss2(self))


class CutLevelBoss(CutScene):
    lines = [
        "",
        "The God In the World.",
        ""
        "Your goal is at hand. Kill or be killed."

        ]

    next_scene = LevelBoss


class LevelBoss2(Level0):

    def __init__(self, prev):
        super(LevelBoss2, self).__init__()
        self.player = prev.player
        self.player.velocity = euclid.Point2(0, 0)
        self.add(prev.player)
        self.start_x = self.player.x
        self.start_y = self.player.y

    def setup_level(self):
        self.goal = Goal()
        self.goal.x, self.goal.y = -300, 300
        self.goal.z = -2
        self.add(self.goal)

        wpu = Boss2()
        wpu.x, wpu.y = 300, 300
        wpu.z = -1
        self.add(wpu)

    def setup_player(self):
        pass

    def goal_reached(self):
        if self.player.killed_her:
            self.director.replace(EndKill(self))
        else:
            self.director.replace(EndNoKill(self))


class CutLevelBoss2(CutScene):
    lines = [
        "",
        "The Genius and the Goddess.",
        "",
        "So, nothing but A Woman's Vengeance after all.",
        "Stay with her,",
        "and write the Foreword to the Supreme Doctrine?",
        "Can you forget about The Human Situation?",
        ]

    next_scene = LevelBoss2


class EndKill(CutScene):
    is_final = True

    lines = [
        "",
        "A Brave New World.",
        "",
        "The age of gods has finished.",
        "Misery remains and hope is lost.",
        "But at least its your doing.",
        "",
        "What would have happened had you chosen differently?",
        ]

    def event(self, ev):
        if self.readkeys:
            if ev.type == KEYDOWN:
                if ev.key == K_SPACE:
                        self.director.replace(Title())


class EndNoKill(CutScene):
    is_final = True

    lines = [
        "",
        "A Brave New World.",
        "",
        "Living with the gods and ruling.",
        "Misery remains and its your doing.",
        "But at least yo are a god.",
        "",
        "What would have happened had you chosen differently?",
        ]

    def event(self, ev):
        if self.readkeys:
            if ev.type == KEYDOWN:
                if ev.key == K_SPACE:
                        self.director.replace(Title())